#!/usr/bin/env python
# encoding: utf-8

"""
funcutils provided some useful functions.
"""

import re
import operator
import inspect


def template_render(template, model):
    """
    A simple template render.
    Example:
    >>> model = {
    ...     "name": "lyc",
    ...     "foo": {
    ...       "bar": "hello"
    ...     }
    ...   }
    >>> template = "{{foo.bar}} {{ name }}"
    >>> template_render(template, model)

    :param template: template string
    :param model: dict model
    """
    parts = []
    for part in re.split(r"{{([\.\w]+?)}}", template):
        parts.append(str(
            part if len(parts) & 1 == 0 else
            reduce(lambda m, k: m[k], part.split("."), model)))
    return "".join(parts)


def encode(s, encoding="utf-8", errors="strict"):
    """
    Auto encode unicode string to bytes.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: bytes
    """
    if isinstance(s, bytes):
        return s

    return s.encode(encoding, errors)


def decode(s, encoding="utf-8", errors="strict"):
    """
    Auto decode bytes to unicode string.
    :param s: string
    :param encoding: bytes encoding(default:utf-8)
    :param errors: different error handling scheme
    :return: unicode
    """
    if isinstance(s, unicode):
        return s

    return s.decode(encoding, errors)


def bytescat(s1, s2, encoding="utf-8"):
    """
    Auto encode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding: bytes encoding(default:utf-8)
    :return: bytes
    """
    return encode(s1, encoding) + encode(s2, encoding)


def strcat(s1, s2, encoding1="utf-8", encoding2=None):
    """
    Auto decode and return s1 + s2
    :param s1: string
    :param s2: string
    :param encoding1: bytes encoding of s1(default:utf-8)
    :param encoding2: bytes encoding of s2(default:same as encoding1)
    :return: bytes
    """
    encoding2 = encoding2 or encoding1
    return decode(s1, encoding1) + decode(s2, encoding2)


def drop_prefix(s, pattern):
    """
    Remove prefix pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.startswith(pattern):
        return s[len(pattern):]
    return s


def drop_postfix(s, pattern):
    """
    Remove postfix pattern of a string.

    :param s: string
    :param pattern: string pattern
    """
    if s.endswith(pattern):
        return s[:-len(pattern)]
    return s


def split_and_strip(val_str, sep=","):
    """
    Simple split val_str by sep and drop strip the space chars for each items.

    :param val_str: string
    :param sep: split separator
    """
    return [
        i for i in (
            i.strip() for i in val_str.split(sep)
        )
        if i
    ]


def left_part_of(txt, sub_txt, n=1):
    """
    Return the left part before the nth of sub_txt appeared in txt.

    :param txt: text
    :param sub_txt: separate text
    :param n: the nth of sub_txt(default:1)
    """
    parts = txt.split(sub_txt)
    return sub_txt.join(parts[:n])


def right_part_of(txt, sub_txt, n=-1):
    """
    Return the right part after the nth of sub_txt appeared in txt.

    :param txt: text
    :param sub_txt: separate text
    :param n: the nth of sub_txt(default:-1)
    """
    parts = txt.split(sub_txt)
    return sub_txt.join(parts[n:])


def is_magic_method(method):
    """
    Check a method if is magic method

    :param method: method of a class
    """
    if not inspect.ismethod(method):
        return False
    func_name = method.im_func.func_name
    if func_name.startswith("__") and func_name.endswith("__"):
        return True
    return False


def set_default_attr(obj, name, value):
    """
    Set attribute to obj if attribute is not exists

    :param obj: object
    :param name: attribure name
    :param value: attribure value
    :return: if is attribute not exists
    """
    if not hasattr(obj, name):
        setattr(obj, name, value)
    return getattr(obj, name, value)


def iter_attrs(
    obj, include_fields=(), exclude_fields=(),
    public_only=True, exclude_methods=True,
):
    """
    iter attributes of obj

    :param obj: object
    :param include_fields: include all this fields if exists
    :param exclude_fields: exclude this fields
    :param public_only: choice public attr only
    :param exclude_methods: exclude methods
    """
    obj_type = type(obj)
    obj_type_name = obj_type.__name__
    for attr, val in inspect.getmembers(obj):
        if attr not in include_fields:
            if (
                attr in exclude_fields
                or attr.startswith("__")
                or attr.startswith("_%s__" % obj_type_name)
                or (public_only and attr.startswith("_"))
                or (exclude_methods and inspect.ismethod(val))
            ):
                continue
        yield attr, val


def filter_n(func_or_none, sequence, n=1):
    """
    Return n items in sequence which match func_or_none.

    :param func_or_none: judge function or None to choice items that are True
    :param sequence: iterable sequence
    :param n: choice n items at most
    """
    func = func_or_none or (lambda x: x)
    result = []
    cnt = 0
    for i in sequence:
        if cnt >= n:
            break
        if func(i):
            result.append(i)
            cnt += 1
    return result



#!/usr/bin/env python
# encoding: utf-8

import os
import codecs
import tempfile
import shutil
from contextlib import contextmanager
import exceptions
from functools import wraps
import hashlib

from ycyc.base import contextutils


def oserror_format(func):
    """
    Decorator that catch WindowsError and reraise it as OSError
    """
    @wraps(func)
    def foo(*args, **kwargs):
        with contextutils.catch(
            getattr(exceptions, "WindowsError", OSError),
            reraise=OSError,
        ):
            return func(*args, **kwargs)
    return foo


@oserror_format
def make_sure_dir_exists(path):
    """
    Create the directories when the path is not exists.

    :param path: path string
    """
    if not os.path.exists(path):
        os.makedirs(path)


@oserror_format
def remove_dir(path, recursion=False):
    """
    Remove the directory.

    :param path: path string
    :param recursion: if delete sub directories and files
    """
    if recursion:
        shutil.rmtree(path)
    else:
        os.removedirs(path)


@oserror_format
def make_sure_dir_empty(path):
    """
    Make sure the directory is empty.

    :param path: path string
    """
    if not os.path.exists(path):
        make_sure_dir_exists(path)
    else:
        for i in os.listdir(path):
            i_path = os.path.join(path, i)
            if os.path.isdir(i_path):
                remove_dir(i_path)
            else:
                os.remove(i_path)


@oserror_format
def make_sure_not_exists(path):
    """
    Make sure the path is not exists.

    :param path: path string
    """
    if not os.path.exists(path):
        return
    if os.path.isdir(path):
        remove_dir(path, recursion=True)
    else:
        os.remove(path)


def safe_open_for_write(fn, encoding="utf-8"):
    """
    Auto create the directories and open a file for write(w+).

    :param fn: file path string
    :param encoding: file encoding
    """
    make_sure_dir_exists(os.path.dirname(fn))
    return codecs.open(fn, "w+", encoding)


def touch_file(path):
    """
    As same as *nix command `touch`

    :param path: path string
    """
    with safe_open_for_write(path):
        pass


def safe_open_for_update(fn, encoding="utf-8"):
    """
    Make sure open a file for update(r+).

    :param fn: file path string
    :param encoding: file encoding
    """
    if not os.path.exists(fn):
        return safe_open_for_write(fn, encoding)
    return codecs.open(fn, "r+", encoding)


def safe_open_for_read(fn, encoding="utf-8"):
    """
    Make sure open a file for read(r).

    :param fn: file path string
    :param encoding: file encoding
    """
    if os.path.exists(fn):
        return codecs.open(fn, "r", encoding)
    return tempfile.TemporaryFile("r")


@contextmanager
@oserror_format
def cd(path):
    """
    Switch to a directory to excute.

    :param path: directory string
    """
    current_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        with contextutils.catch():
            os.chdir(current_dir)


def bytes_from(path):
    """
    Return the bytes from a file.

    :param path: file path string
    """
    with open(path, "rb") as fp:
        return fp.read()


def text_from(path, encoding="utf-8"):
    """
    Return the text from a file.

    :param path: file path string
    :param encoding: file encoding
    """
    with codecs.open(path, "r", encoding) as fp:
        return fp.read()


def hash_of(path, hash_name):
    """
    Return the hash of a file

    :param path: file path string
    :param hash_name: hash algorithm name(md5/sha1/sha256/sha512)
    """
    hash = hashlib.new(hash_name, bytes_from(path))
    return hash.hexdigest()


def sha1_of(path):
    """
    Return the sha1 hash of a file

    :param path: file path string
    """
    return hash_of(path, "sha1")


def md5_of(path):
    """
    Return the md5 hash of a file

    :param path: file path string
    """
    return hash_of(path, "md5")



#!/usr/bin/env python
# encoding: utf-8

import json
from functools import wraps

from ycyc.base.decoratorutils import cachedproperty


class RichTypeJSONEncoder(object):
    def __init__(self, type_mapping):
        self.__type_mapping = type_mapping

    @cachedproperty
    def encoder_factory(self):
        type_mapping = self.__type_mapping

        class JsonEncoder(json.JSONEncoder):
            def __init__(self, *args, **kwg):
                self.base = super(JsonEncoder, self)
                self.base.__init__(*args, **kwg)

            def default(self, data):  # pylint: disable=method-hidden
                factory = type_mapping.get(type(data))
                if factory:
                    return factory(data)
                try:
                    return self.base.default(data)
                except TypeError:
                    for typ, fcty in type_mapping.items():
                        if isinstance(data, typ):
                            return fcty(data)
                    raise
        return JsonEncoder

    @classmethod
    def default_type_mapping(cls):
        import datetime
        import decimal
        import collections
        import types

        return {
            (
                set, frozenset,
                types.GeneratorType,
                types.XRangeType,
                collections.deque,
            ): list,
            (
                collections.Mapping,
                collections.defaultdict,
                collections.Counter,
                collections.OrderedDict,
            ): dict,

            datetime.date: datetime.date.isoformat,
            datetime.datetime: datetime.datetime.isoformat,
            decimal.Decimal: float,
        }


def dumps(obj):
    encoder = RichTypeJSONEncoder(RichTypeJSONEncoder.default_type_mapping())
    return json.dumps(obj, cls=encoder.encoder_factory)



#!/usr/bin/env python
# encoding: utf-8

__author__ = 'Liu Yicong'
__email__ = 'imyikong@gmail.com'

from contextlib import contextmanager, nested

import mock


@contextmanager
def mock_patches(*patches, **named_patches):
    """
    A context manager to help create mock patches.
    >>> with mock_patches("package.module.cls", cls2="package.cls") as mocks:
    ...     mocks.cls() #=> package.module.cls
    ...     mocks.cls2() #=> package.cls
    """
    attrs = list(i.split(".")[-1] for i in patches)
    attrs.extend(named_patches.keys())
    patches = list(patches)
    patches.extend(named_patches.values())
    mocks = mock.Mock()

    with nested(*map(mock.patch, patches)) as mp:
        for k, m in zip(attrs, mp):
            setattr(mocks, k, m)

        yield mocks



#!/usr/bin/env python
# encoding: utf-8

import collections
from ycyc.base.iterutils import getfirst
from ycyc.collections.tagmaps import TagMaps

__all__ = TagMaps()


@__all__.register("GenData")
class GenData(collections.Mapping):
    def __init__(self, base, extend=None):
        self.base_data = base
        self.extend_data = extend
        self.real_keys = base.viewkeys() | extend.viewkeys()

    def __getitem__(self, key):
        if key in self.extend_data:
            return self.extend_data[key]
        return self.base_data[key]

    def __iter__(self):
        return iter(self.real_keys)

    def __len__(self):
        return len(self.real_keys)


@__all__.register("TxtGenerator")
class TxtGenerator(object):
    ITEM_SEP = ""
    NEWLINE_SEP = "\n"

    def __init__(self, data=None):
        self.output_lst = collections.deque()
        self.data = data or {}

    def write(self, txt, **extend_data):
        data = GenData(self.data, extend_data)
        self.output_lst.append(txt.format(**data))

    def writeline(self, txt, **extend_data):
        self.write(txt + self.NEWLINE_SEP, **extend_data)

    def prepare(self):
        if len(self.output_lst) > 1:
            txt = self.ITEM_SEP.join(self.output_lst)
            self.output_lst.clear()
            self.output_lst.append(txt)

    def getval(self):
        self.prepare()
        return getfirst(self.output_lst)

    def clear(self):
        self.output_lst.clear()

    def pop(self):
        self.output_lst.pop()



#!/usr/bin/env python
# encoding: utf-8

import ast
import importlib
import inspect


class Analysis(ast.NodeTransformer):
    def __init__(self, paths, recursion):
        self.modules = list()
        self.paths = list(paths)
        self.recursion = recursion

    def add_module(self, module):
        if module and module not in self.modules:
            self.modules.append(module)
            if self.recursion:
                try:
                    path = inspect.getsourcefile(importlib.import_module(module))
                    if path:
                        self.paths.append(path)
                except:
                    pass

    def visit_Import(self, node):
        for i in node.names:
            self.add_module(i.name)

    def visit_ImportFrom(self, node):
        self.add_module(node.module)

    def analysis(self):
        for p in self.paths:
            try:
                with open(p, "rt") as fp:
                    self.visit(ast.parse(fp.read(), p))
            except:
                pass
        return tuple(self.modules)

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("paths", nargs="+")
    parser.add_argument("-r", "--recursion", action="store_true", default=False)
    args = parser.parse_args()

    analysisor = Analysis(args.paths, args.recursion)
    for m in analysisor.analysis():
        print m



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.collections import tagmaps


class TestTagMaps(TestCase):
    def test_uasge(self):
        maps = tagmaps.TagMaps()

        @maps.register("add")
        def add(x, y):
            return x + y

        @maps.register("sub")
        def sub(x, y):
            return x - y

        self.assertEqual(maps["add"](1, 2), add(1, 2))
        self.assertEqual(maps["sub"](4, 5), sub(4, 5))

        with self.assertRaises(KeyError):
            self.assertEqual(maps["noexist"](6, 7), None)

        @maps.register(maps.DefaultKey)
        def default(x, y):
            return None

        self.assertEqual(maps["noexist"](8, 9), default(8, 9))

        self.assertListEqual(list(maps), ["", "add", "sub"])
        self.assertEqual(list(maps)[0], "")
        self.assertEqual(list(maps)[1], "add")
        self.assertEqual(list(maps)[2], "sub")



#!/usr/bin/env python
# encoding: utf-8

"""
iterutils provided some utils to operate a iterable or indexable object.
"""

import itertools
import collections
import logging

logger = logging


def getitems(obj, items, default=None):
    """
    Get items from obj
    :param obj: object
    :param items: string or iterable object
    :param default: default value
    :return: obj item value if existed, otherwise default value
    """
    if isinstance(items, (str, unicode)):
        items = (items,)

    try:
        return reduce(lambda x, i: x[i], items, obj)
    except (IndexError, KeyError):
        return default


def getattrs(obj, attrs, default=None):
    """
    Get attrs from obj
    :param obj: object
    :param attrs: string or iterable object
    :param default: default value
    :return: obj attr value if existed, otherwise default value
    """
    if isinstance(attrs, (str, unicode)):
        attrs = (attrs,)

    try:
        return reduce(lambda x, n: getattr(x, n), attrs, obj)
    except AttributeError:
        return default


def iterable(obj):
    """
    Check obj if is iterable
    :param obj: object
    :return: boolean
    """
    return hasattr(obj, "__iter__")


def getnext(iterator, default=None):
    """
    Get next item of iterable object
    :param iterator: object
    :param default: default value
    :return: next item of iterator till the end, otherwise default value
    """
    try:
        return next(iterator)
    except StopIteration:
        return default


def getfirst(obj, default=None):
    """
    Get first item of iterable object
    :param obj: object
    :param default: default value
    :return: first item of obj if obj is not empty, otherwise default value
    """
    if not obj:
        return default
    return getnext(iter(obj), default)


def groupby(iterable, keyfunc):
    """
    Group iterable by keyfunc
    :param iterable: iterable obj
    :param keyfunc: group by keyfunc
    :return: dict {<keyfunc result>: [item1, item2]}
    """
    grouper = itertools.groupby(iterable, keyfunc)
    group_dict = collections.defaultdict(list)
    for k, r in grouper:
        group = group_dict[k]
        group.extend(r)
    return dict(group_dict)


def mkparts(sequence, indices=None):
    """
    Make some parts from sequence by indices
    :param sequence: indexable object
    :param indices: index list
    :return: [seq_part1, seq_part2, ...]
    """
    indices = indices or [1]
    result_list = collections.deque()
    start = 0
    seq_len = len(sequence)
    for end in indices:
        if end < 0:
            end = seq_len + end
        if end < start:
            raise ValueError("end index is less than start index")
        result_list.append(sequence[start:end])
        start = end
    result_list.append(sequence[start:])
    return result_list


def get_single_item(obj, default=None, logger=logger):
    """
    Get the single item from obj, if obj has more than 1 items, it
    will log a warning record.
    :param obj: iterable obj
    :param default: default value if obj is empty
    :return: obj first item
    """
    if obj and len(obj) > 1:
        logger.warning(
            "iterable object has more than 1 items, %s actually.", len(obj)
        )
    return getfirst(obj, default)


def dict_merge(dicts):
    """
    Merge dict list to a dict object
    :param dicts: dict list
    :return: dict
    """
    if not dicts:
        raise IndexError()

    result = {}

    for d in dicts:
        for k, v in d.items():
            if k not in result:
                result[k] = v

    return result


def flatten(sequence_list, cls=list):
    """
    Flatten one level of nesting
    :param sequence_list: list of sequence
    :param cls: create instance of cls by flatten_gen
    :return: cls instance or generator
    """
    flatten_gen = itertools.chain.from_iterable(sequence_list)
    return cls(flatten_gen) if cls else flatten_gen



#!/usr/bin/env python
# encoding: utf-8

import inspect

from ycyc.base.decoratorutils import allow_unbound_method


def init_if_is_class(cls_or_ins):
    if inspect.isclass(cls_or_ins):
        cls_or_ins = cls_or_ins()
    return cls_or_ins


class Rescue(object):
    @allow_unbound_method
    def catch(cls_or_ins, value, exception=Exception):  # pylint: disable=E0213
        self = init_if_is_class(cls_or_ins)
        self.exceptions[exception] = value
        return self

    @allow_unbound_method
    def meet(cls_or_ins, value, wish=None):  # pylint: disable=E0213
        self = init_if_is_class(cls_or_ins)
        self.conversions[wish] = value
        return self

    def __init__(self):
        self.exceptions = {}
        self.conversions = {}

    def call(self, func, *args, **kwg):
        try:
            result = func(*args, **kwg)
            return self.conversions.get(result, result)
        except Exception as err:
            err_typ = type(err)
            if err_typ in self.exceptions:
                return self.exceptions[err_typ]
            for types, val in self.exceptions.items():
                if isinstance(err, types):
                    return val



#!/usr/bin/env python
# encoding: utf-8

from contextlib import contextmanager
import six
import logging

logger = logging.getLogger(__name__)


@contextmanager
def catch(errors=Exception, reraise=None, callback=logger.warning):
    """
    A context manager to catch the exceptions raised from block.
    :param errors: exception or exceptions tuple
    :param reraise: reraise a new exception from catched exception(default: None)
    :param callback: callback when catched a exception(default: logger.warning)
    """
    try:
        yield
    except errors as err:
        if callback:
            callback(err)
        if reraise:
            six.raise_from(reraise, err)


@contextmanager
def subprocessor(*args, **kwg):
    """
    With new subprocess call,terminate it when exit context
    :param args: argument pass to subprocess.Popen
    :param kwg: key word argument pass to subprocess.Popen
    """
    from subprocess import Popen
    processor = Popen(*args, **kwg)
    try:
        yield processor
    finally:
        if processor.poll() is None:
            processor.terminate()
            processor.wait()


@contextmanager
def timeout(seconds, interval=None, ticks=None):
    """
    Send KeyboardInterrupt to main thread to terminate it and
    convert it as RuntimeError if timeout.
    :param seconds: timeout seconds
    :param interval: poll interval
    :param ticks: CPU-Bound thread check interval
    """
    import threading
    import time
    import os
    import signal
    import sys

    signal_finished = False
    start = time.time()
    interval = interval or 0.1
    old_ticks = sys.getcheckinterval()

    cur_thread = threading.current_thread()
    assert cur_thread.name == "MainThread"

    def poll_signal():
        now = time.time()
        while not signal_finished and now - start < seconds:
            time.sleep(interval)
            now = time.time()
        if not signal_finished:
            os.kill(os.getpid(), signal.SIGINT)

    if seconds > 0:
        poll_thread = threading.Thread(target=poll_signal)
        poll_thread.daemon = True
        poll_thread.start()
    else:
        poll_thread = None

    if ticks is not None:
        sys.setcheckinterval(ticks)
    try:
        yield
    except KeyboardInterrupt:
        now = time.time()
        if now - start >= seconds > 0:
            raise RuntimeError("timeout")
        raise
    finally:
        signal_finished = True
        if ticks is not None:
            sys.setcheckinterval(old_ticks)
        if poll_thread and poll_thread.is_alive():
            poll_thread.join()


@contextmanager
def nothing(*args, **kwg):
    """
    Just a place holder.
    """
    yield


@contextmanager
def atlast(func, force=False):
    """
    Run func at last
    """
    try:
        yield
        func()
    except Exception:
        if force:
            func()
        raise



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

import mock

from ycyc.base.allowfail import AllowFail


class TestUsage(TestCase):
    @mock.patch("ycyc.base.allowfail.AllowFail.on_error")
    def test_decorator(self, magic_mock):
        @AllowFail("test func")
        def func():
            return 1

        result, exception = func()
        self.assertIsNone(exception)
        self.assertEqual(result, 1)
        self.assertIsNone(magic_mock.call_args)

    @mock.patch("ycyc.base.allowfail.AllowFail.on_error")
    def test_decorator_err(self, magic_mock):
        @AllowFail("test valueerror")
        def valueerror():
            raise ValueError("test exception")

        result, exception = valueerror()
        self.assertIsInstance(exception, ValueError)
        self.assertEqual(exception.message, "test exception")
        args = magic_mock.call_args[0]
        self.assertEqual(args[0], "test valueerror")
        self.assertIs(args[1], exception)

    @mock.patch("ycyc.base.allowfail.AllowFail.on_error")
    def test_contextmanager(self, magic_mock):
        with AllowFail("test contextmanager"):
            pass

        self.assertIsNone(magic_mock.call_args)

    @mock.patch("ycyc.base.allowfail.AllowFail.on_error")
    def test_contextmanager_err(self, magic_mock):
        with AllowFail("test contextmanager"):
            raise ValueError()

        args = magic_mock.call_args[0]
        self.assertEqual(args[0], "test contextmanager")
        self.assertTrue(isinstance(args[1], ValueError))



#!/usr/bin/env python
# encoding: utf-8

"""
A adapter tool module
"""

import collections
import sys
import functools


class ObjAsDictAdapter(collections.Mapping):
    """
    Provided a adapter that allow visit a object as dict
    """
    def __init__(self, obj):
        self.__object = obj

    def __getitem__(self, name):
        try:
            return getattr(self.__object, name)
        except AttributeError:
            raise KeyError(name)

    def __setitem__(self, name, value):
        setattr(self.__object, name, value)

    def __iter__(self):
        return iter(dir(self.__object))

    def __len__(self):
        return len(dir(self.__object))


def main_entry(main):
    """
    Call function `main` immediately if is in __main__ and exit
    """
    @functools.wraps(main)
    def wraped_main(*args, **kwg):
        result = main(*args, **kwg)
        return int(0 if result is None else result)

    if main.__module__ == "__main__":
        sys.exit(
            wraped_main()
            if main.func_code.co_argcount == 0
            else wraped_main(sys.argv)
        )

    return wraped_main



#!/usr/bin/env python
# encoding: utf-8

import six
from collections import OrderedDict
from itertools import chain
import logging

from ycyc.base.contextutils import catch

logger = logging.getLogger(__name__)


class RequireFieldsMissError(Exception):
    pass


class NamedDictMeta(type):
    def __new__(cls, name, base, attrs):
        fields = attrs.pop("__Fields__", {})
        attrs["__Fields__"] = OrderedDict(fields)

        requires = attrs.pop("__Requires__", ())
        attrs["__Requires__"] = tuple(requires)

        return type.__new__(cls, name, base, attrs)


class NamedDict(six.with_metaclass(NamedDictMeta, dict)):
    def __init__(self, *args, **kwg):
        requires = self.__Requires__
        fields = self.__Fields__.copy()
        field_keys = fields.viewkeys()
        real_keys = field_keys | set(requires)

        if kwg:
            kwg_keys = kwg.viewkeys()
            undefined_keys = kwg_keys - real_keys
            if undefined_keys:
                logger.warning("undefined keys: %s", undefined_keys)

            fields.update({
                k: kwg[k]
                for k in real_keys & kwg_keys
            })
        if args:
            i_args = iter(args)
            fields.update({
                k: v
                for k, v in zip(chain(requires, field_keys), i_args)
            })

            remain_args = list(i_args)
            if remain_args:
                logger.warning("not acceptable arguments: %s", remain_args)

        missed_fields = set(requires) - set(fields.keys())
        if missed_fields:
            logger.debug("missed_fields: %s", missed_fields)
            raise RequireFieldsMissError(
                "fields: [%s]" % ", ".join(missed_fields)
            )

        super(NamedDict, self).__init__(fields)

    def __getattr__(self, name):
        with catch(KeyError, reraise=AttributeError):
            return self[name]

    def __setattr__(self, name, key):
        self[name] = key


def namedict(name, requires=(), fields=None):
    """
    Return a subclass of dict with named fields

    :param fields: option fields
    :param requires: requires fields
    """
    return type(name, (NamedDict,), {
        "__Fields__": fields or {},
        "__Requires__": requires,
    })



#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase
from collections import namedtuple
import types

from ycyc.base.iterutils import (
    getitems, getattrs, iterable, getnext, getfirst, mkparts,
    get_single_item, dict_merge, flatten
)

import mock


class Testgetitems(TestCase):
    def test_usage(self):
        model = {
            "key": "test",
            "value": {
                "items": [
                    {
                        "name": "a",
                        "value": 1
                    },
                    {
                        "name": "b",
                        "value": 2
                    }
                ]
            }
        }

        self.assertEqual(getitems(model, "key"), "test")
        self.assertEqual(getitems(model, ["value", "items", 0, "name"]), "a")
        self.assertEqual(getitems(model, ["value", "items", 1, "value"]), 2)
        self.assertEqual(getitems(model, ["value", "items", 2, "value"]), None)
        self.assertEqual(getitems(model, ["value", "items", 3, "value"], 4), 4)


class Testgetattrs(TestCase):
    Pair = namedtuple("Pair", ["key", "value"])

    def test_usage(self):
        Pair = self.Pair
        model = Pair(
            key="test",
            value=Pair(key="a", value=1)
        )

        self.assertEqual(getattrs(model, "key"), "test")
        self.assertEqual(getattrs(model, ["value", "key"]), "a")
        self.assertEqual(getattrs(model, ["value", "value"]), 1)
        self.assertEqual(getattrs(model, ["value", "items"]), None)
        self.assertEqual(getattrs(model, ["value", "items"], 4), 4)


class TestIterObj(TestCase):
    def test_iterable(self):
        self.assertFalse(iterable(object()))
        self.assertTrue(iterable(list()))
        self.assertTrue(iterable(tuple()))
        self.assertTrue(iterable(set()))
        self.assertTrue(iterable(dict()))
        self.assertTrue(iterable(i for i in range(0)))

    def test_getfirst(self):
        self.assertEqual(getfirst(range(5)), 0)
        self.assertEqual(getfirst({"a": 1}), "a")

        gen = (i for i in range(1))
        self.assertEqual(getfirst(gen), 0)
        with self.assertRaises(StopIteration):
            next(gen)

        self.assertEqual(getfirst(None), None)
        self.assertEqual(getfirst([]), None)
        self.assertEqual(getfirst([], 0), 0)

    def test_getnext(self):
        gen = (i for i in range(2))
        self.assertEqual(getnext(gen), 0)
        self.assertEqual(getnext(gen), 1)
        self.assertEqual(getnext(gen), None)
        self.assertEqual(getnext(gen, 2), 2)


class TestMakeParts(TestCase):
    def test_usage(self):
        lst = list(range(10))
        part1, part2, part3, part4 = mkparts(lst, [1, 3, 6])
        self.assertListEqual(part1, [0])
        self.assertListEqual(part2, [1, 2])
        self.assertListEqual(part3, [3, 4, 5])
        self.assertListEqual(part4, [6, 7, 8, 9])

        head, middle, tail = mkparts(lst, [1, -1])
        self.assertEqual(head, [0])
        self.assertEqual(middle, [1, 2, 3, 4, 5, 6, 7, 8])
        self.assertEqual(tail, [9])

    def test_special_val(self):
        empty_lst = []
        part1, part2, part3 = mkparts(empty_lst, [2, 4])
        self.assertListEqual(part1, [])
        self.assertListEqual(part2, [])
        self.assertListEqual(part3, [])

        lst = list(range(3))
        part1, part2 = mkparts(lst, [])
        self.assertListEqual(part1, [0])
        self.assertListEqual(part2, [1, 2])
        part1, part2 = mkparts(lst, [1])
        self.assertListEqual(part1, [0])
        self.assertListEqual(part2, [1, 2])

        part1, part2 = mkparts(lst, [4])
        self.assertListEqual(part1, [0, 1, 2])
        self.assertListEqual(part2, [])

        part1, part2, part3 = mkparts(lst, [1, 2])
        self.assertListEqual(part1, [0])
        self.assertListEqual(part2, [1])
        self.assertListEqual(part3, [2])

        part1, part2, part3 = mkparts(lst, [1, 1])
        self.assertListEqual(part1, [0])
        self.assertListEqual(part2, [])
        self.assertListEqual(part3, [1, 2])

        with self.assertRaisesRegexp(ValueError, "end index is less than start index"):
            part1, part2, part3 = mkparts(lst, [2, 1])

        with self.assertRaisesRegexp(ValueError, "end index is less than start index"):
            part1, part2, part3 = mkparts(lst, [2, -2])


class TestGetSingleItem(TestCase):
    def test_usage(self):
        mock_logger = mock.MagicMock()

        self.assertEqual(get_single_item([], logger=mock_logger), None)
        self.assertEqual(mock_logger.warning.call_count, 0)
        self.assertEqual(get_single_item([], 1, logger=mock_logger), 1)
        self.assertEqual(mock_logger.warning.call_count, 0)
        self.assertEqual(get_single_item([1], logger=mock_logger), 1)
        self.assertEqual(mock_logger.warning.call_count, 0)
        self.assertEqual(get_single_item([1, 2], logger=mock_logger), 1)
        self.assertEqual(mock_logger.warning.call_count, 1)


class TestDictMerge(TestCase):
    def test_usage(self):
        with self.assertRaises(IndexError):
            dict_merge([])

        self.assertDictEqual(
            dict_merge([
                {1: 2, 2: 3},
                {2: 4, 3: 4},
                {4: 5},
            ]),
            {1: 2, 2: 3, 3: 4, 4: 5}
        )


class TestFlatten(TestCase):
    def test_usage(self):
        seq_list = [range(0, 3), range(3, 7), (i for i in range(7, 9))]
        self.assertListEqual(flatten(seq_list, list), range(0, 9))


#!/usr/bin/env python
# encoding: utf-8

from unittest import TestCase

from ycyc.base import funcutils


class TestCoding(TestCase):
    b_en = b"test"
    b_ch = u"测试".encode("utf-8")
    u_en = u"test"
    u_ch = u"测试"

    def test_encode(self):
        self.assertEqual(self.b_en, funcutils.encode(self.u_en, "utf-8"))
        self.assertEqual(self.b_ch, funcutils.encode(self.u_ch, "utf-8"))
        self.assertEqual(self.b_en, funcutils.encode(self.b_en, "utf-8"))
        self.assertEqual(self.b_ch, funcutils.encode(self.b_ch, "utf-8"))

    def test_decode(self):
        self.assertEqual(self.u_en, funcutils.decode(self.u_en, "utf-8"))
        self.assertEqual(self.u_ch, funcutils.decode(self.u_ch, "utf-8"))
        self.assertEqual(self.u_en, funcutils.decode(self.b_en, "utf-8"))
        self.assertEqual(self.u_ch, funcutils.decode(self.b_ch, "utf-8"))

    def test_bytescat(self):
        target = u"test测试".encode("utf-8")
        self.assertEqual(funcutils.bytescat(self.b_en, self.b_ch), target)
        self.assertEqual(funcutils.bytescat(self.u_en, self.u_ch), target)
        self.assertEqual(funcutils.bytescat(self.b_en, self.u_ch), target)
        self.assertEqual(funcutils.bytescat(self.u_en, self.b_ch), target)

        target = u"测试测试".encode("utf-8")
        self.assertEqual(funcutils.bytescat(self.b_ch, self.b_ch), target)
        self.assertEqual(funcutils.bytescat(self.u_ch, self.u_ch), target)
        self.assertEqual(funcutils.bytescat(self.b_ch, self.u_ch), target)
        self.assertEqual(funcutils.bytescat(self.u_ch, self.b_ch), target)

    def test_strcat(self):
        target = u"test测试"
        self.assertEqual(funcutils.strcat(self.b_en, self.b_ch), target)
        self.assertEqual(funcutils.strcat(self.u_en, self.u_ch), target)
        self.assertEqual(funcutils.strcat(self.b_en, self.u_ch), target)
        self.assertEqual(funcutils.strcat(self.u_en, self.b_ch), target)

        target = u"测试测试"
        self.assertEqual(funcutils.strcat(self.b_ch, self.b_ch), target)
        self.assertEqual(funcutils.strcat(self.u_ch, self.u_ch), target)
        self.assertEqual(funcutils.strcat(self.b_ch, self.u_ch), target)
        self.assertEqual(funcutils.strcat(self.u_ch, self.b_ch), target)

        gbk_ch = u"测试".encode("gbk")
        self.assertEqual(
            funcutils.strcat(gbk_ch, self.b_ch, "gbk", "utf-8"), target)
        self.assertEqual(
            funcutils.strcat(gbk_ch, self.u_ch, "gbk", "utf-8"), target)
        self.assertEqual(
            funcutils.strcat(self.b_ch, gbk_ch, "utf-8", "gbk"), target)
        self.assertEqual(
            funcutils.strcat(self.u_ch, gbk_ch, "utf-8", "gbk"), target)


class TestStringUtils(TestCase):
    def test_drop_prefix(self):
        self.assertEqual(
            funcutils.drop_prefix("root/file.txt", "root/"),
            "file.txt"
        )
        self.assertEqual(
            funcutils.drop_prefix("dir/file.txt", "root/"),
            "dir/file.txt"
        )

    def test_drop_postfix(self):
        self.assertEqual(
            funcutils.drop_postfix("file.txt", ".txt"),
            "file"
        )
        self.assertEqual(
            funcutils.drop_postfix("file.tmp", ".txt"),
            "file.tmp"
        )

    def test_split_and_strip(self):
        self.assertListEqual(
            funcutils.split_and_strip("a,b ,c, d , e"),
            ["a", "b", "c", "d", "e"]
        )

    def test_left_part_of(self):
        txt = "xxx.avi.part1.rar"
        self.assertEqual(funcutils.left_part_of(txt, "."), "xxx")
        self.assertEqual(funcutils.left_part_of(txt, ".", 0), "")
        self.assertEqual(funcutils.left_part_of(txt, ".", 1), "xxx")
        self.assertEqual(funcutils.left_part_of(txt, ".", 2), "xxx.avi")
        self.assertEqual(funcutils.left_part_of(txt, ".", -1), "xxx.avi.part1")
        self.assertEqual(funcutils.left_part_of(txt, ".", -2), "xxx.avi")

    def test_right_part_of(self):
        txt = "xxx.avi.part1.rar"
        self.assertEqual(funcutils.right_part_of(txt, "."), "rar")
        self.assertEqual(funcutils.right_part_of(txt, ".", 0), txt)
        self.assertEqual(funcutils.right_part_of(txt, ".", -1), "rar")
        self.assertEqual(funcutils.right_part_of(txt, ".", -2), "part1.rar")
        self.assertEqual(funcutils.right_part_of(txt, ".", 1), "avi.part1.rar")
        self.assertEqual(funcutils.right_part_of(txt, ".", 2), "part1.rar")


class TestObjectHelper(TestCase):
    def test_is_magic_method(self):
        class Foo(object):
            def __str__(self):
                return "foo"

        foo = Foo()
        self.assertTrue(funcutils.is_magic_method(Foo.__str__))
        self.assertTrue(funcutils.is_magic_method(foo.__str__))

    def test_set_default_attr(self):
        class Foo(object):
            def __str__(self):
                return "foo"

        foo = Foo()
        self.assertFalse(hasattr(foo, "foobarbaz"))
        self.assertTrue(funcutils.set_default_attr(
            foo, "foobarbaz", True
        ))
        self.assertTrue(foo.foobarbaz)
        self.assertTrue(funcutils.set_default_attr(
            foo, "foobarbaz", False
        ))
        self.assertTrue(foo.foobarbaz)

    def test_iter_attrs(self):
        class Foo(object):
            Public = 0
            _Protected = 1
            __Privated = 2

            def __init__(self):
                self.public = 4
                self._protected = 5
                self.__privated = 6

            def get_privated(self):
                return self.__privated

            @property
            def protected(self):
                return self._protected

        foo = Foo()

        self.assertDictEqual(
            dict(funcutils.iter_attrs(foo)),
            {
                "Public": 0, "public": 4, "protected": 5,
            }
        )
        self.assertDictEqual(
            dict(funcutils.iter_attrs(foo, public_only=False)),
            {
                "Public": 0, "_Protected": 1, "public": 4,
                "protected": 5, "_protected": 5,
            }
        )
        self.assertDictEqual(
            dict(funcutils.iter_attrs(foo, include_fields={"_Protected"})),
            {
                "Public": 0, "_Protected": 1, "public": 4,
                "protected": 5,
            }
        )
        self.assertDictEqual(
            dict(funcutils.iter_attrs(foo, exclude_fields={"protected"})),
            {
                "Public": 0, "public": 4,
            }
        )
        result = dict(funcutils.iter_attrs(foo, exclude_methods=False))
        self.assertEqual(result["get_privated"], foo.get_privated)

        self.assertIn(
            "_Protected",
            dict(funcutils.iter_attrs(
                foo,
                include_fields={"_Protected"},
                exclude_fields={"_Protected"}
            ))
        )


class TestFilterN(TestCase):
    def test_filter_n(self):
        self.assertListEqual(
            [1, 3],
            funcutils.filter_n(lambda x: x % 2, range(10), 2)
        )
        self.assertListEqual(
            [1],
            funcutils.filter_n(None, range(10), 1)
        )




#!/usr/bin/env python
# encoding: utf-8

from collections import namedtuple
import smtplib
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.header import Header
import os
import six

from ycyc.base.typeutils import SimpleExceptions, constants
from ycyc.base.filetools import bytes_from

SimpleExceptions = SimpleExceptions()
EMailSendFailed = SimpleExceptions.EMailSendFailed

Attachments = namedtuple("Attachments", ["name", "path"])


class EMail(object):
    SubType = constants(
        Plain="plain",
        HTML="html",
    )

    def __init__(
        self, server, sender=None, receiver=None, subject="", content="",
        subtype="plain", charset="utf-8",
    ):
        if isinstance(receiver, basestring):
            receiver = [receiver]

        self.server = server
        self.sender = sender
        self.receiver = receiver or []
        self.subject = subject
        self.content = content
        self.subtype = subtype
        self.charset = charset
        self.attachment_paths = []

    @property
    def mail(self):
        mail = MIMEMultipart()
        mail['Subject'] = Header(self.subject, self.charset)
        mail['From'] = self.sender
        mail['To'] = ",".join(self.receiver)
        mail.attach(MIMEText(self.content, self.subtype, self.charset))

        for i in self.attachment_paths:
            attachment = MIMEApplication(bytes_from(i.path))
            attachment.add_header(
                "Content-Disposition",
                "attachment",
                filename=i.name.strip(),
            )
            mail.attach(attachment)
        return mail

    def send(self, passwd=None, user=None):
        server = smtplib.SMTP(self.server)
        if passwd:
            user = user or self.sender
            server.login(user, passwd)

        try:
            server.sendmail(self.sender, self.receiver, self.mail.as_string())
        except smtplib.SMTPRecipientsRefused as e:
            six.raise_from(EMailSendFailed, e)
        finally:
            server.close()

    def add_attachment(self, path, name=None):
        if not name:
            _, name = os.path.split(path)
        self.attachment_paths.append(Attachments(name, path))



#!/usr/bin/env python
# encoding: utf-8

from ycyc.base import lazyutils

from unittest import TestCase


class TestLazyEnv(TestCase):
    def test_attr_usage(self):
        env = lazyutils.LazyEnv()
        env.a = lambda: global_a

        with self.assertRaisesRegexp(NameError, r"\bglobal_a\b"):
            env.a

        global_a = []
        self.assertIs(env.a, global_a)
        global_a = {}
        self.assertIsNot(env.a, global_a)

        env.a = global_a
        self.assertIs(env.a, global_a)

        global_a = ()
        env.a = lambda: global_a
        self.assertIs(env.a, global_a)

    def test_item_usage(self):
        import types

        env = lazyutils.LazyEnv()
        env["a"] = lambda: global_a

        self.assertIsInstance(env.a, types.LambdaType)
        self.assertIsInstance(env["a"], types.LambdaType)

        env.a = lambda: global_a
        with self.assertRaisesRegexp(NameError, r"\bglobal_a\b"):
            env.a

        global_a = []
        self.assertIs(env.a, global_a)



#!/usr/bin/env python
# encoding: utf-8

import argparse
import os
import textwrap
import six

from ycyc.base.funcutils import drop_prefix
from ycyc.base.adapter import main_entry


class FileExisted(Exception):
    pass


def is_script_file(file):
    return os.path.isfile(file) and file.endswith(".py")


def get_all_scripts(path, recursion=0):
    if is_script_file(path):
        return [path]

    results = []
    if os.path.isdir(path):
        for f in os.listdir(path):
            f_path = os.path.join(path, f)
            if is_script_file(f_path):
                results.append(f_path)
            elif recursion:
                results.extend(get_all_scripts(f_path, recursion - 1))
    return results


def mk_unittest_script(script_relative_path, root):
    path, script = os.path.split(script_relative_path)
    real_dir = os.path.join(root, path)
    real_path = os.path.join(real_dir, "test_%s" % script)

    if os.path.exists(real_path):
        raise FileExisted(real_path)

    if not os.path.exists(real_dir):
        os.makedirs(real_dir)
    with open(real_path, "wt") as fp:
        fp.write(textwrap.dedent('''
        #!/usr/bin/env python
        # encoding: utf-8

        from unittest import TestCase


        class Test1(TestCase):
            pass
        ''').lstrip())


@main_entry
def main():
    working_dir = os.getcwd()
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("path")
    arg_parser.add_argument(
        "-r", "--recursion", default=0, type=int,
    )
    arg_parser.add_argument(
        "-a", "--anchor", default=working_dir,
    )
    arg_parser.add_argument(
        "-t", "--target",
        default=os.path.join(working_dir, "tests"),
    )
    args = arg_parser.parse_args()

    for p in get_all_scripts(args.path, args.recursion):
        dir_path, file = os.path.split(p)
        if file.startswith("__"):
            continue
        try:
            if args.anchor:
                p = drop_prefix(p, args.anchor)
            mk_unittest_script(p, args.target)
        except FileExisted as err:
            six.print_("file is existed:", err)
